/*
' PTLens, Copyright (C) 2004, Thomas Niemann
' email: thomasn@epaperpress.com, subject=ptlens
'
' This program is free software; you can redistribute
' it and/or modify it under the terms of the GNU
' General Public License as published by the Free
' Software Foundation; either version 2 of the
' License, or (at your option) any later version.
'
' This program is distributed in the hope that it will
' be useful, but WITHOUT ANY WARRANTY; without even
' the implied warranty of MERCHANTABILITY or FITNESS
' FOR A PARTICULAR PURPOSE. See the GNU General Public
' License for more details.
'
' You should have received a copy of the GNU General
' Public License along with this program; if not,
' write to the Free Software Foundation, Inc., 675
' Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include "../include/clens.h"



/** \file modDB.c
 * \brief Functions to read lens profile information from profile
 * files.
 * \author Tim Jacobs (ported from Thoms Niemann's PTLens code)
 * \date 2004.11.13
 *
 * Full description here.
 */


extern prefType pref;



/** \brief Points to the first camera in the linked list of 
 * supported cameras.
 */
CamNode *pCamHdr=NULL;
/** \brief Points to the first lens in the linked list of 
 * supported lenses.
 */
LnsNode *pLnsHdr=NULL;
/** \brief Current camera being read in from profiles
 */
CamNode pCurCam;
/** \brief Current lens being read in from profiles
 */
LnsNode pCurLns;
/** \brief Array of lens correction coefficient data types
 */
CoefType coef[10000];
/** \brief Next available slot in coef[]. Holds the total number of lens
 * correction coefficients after all profiles are read.
 */
long coefIndex;
/** \brief Array to hold the filenames contained in profile.txt
 */
char *fileList[500];
/** \brief Next available slot in fileList[].
 */
long fileIndex;
/** \brief Current profile file name being read from
 */
char *fileName;
/** \brief Holds the value of the current line of being read in profile.txt.
 */
int line;
/** \brief String to hold the right hand side of a parsed line.
 */
char rhs[256];
/** \brief State value for parsing profile files
 */
const long PROCESS_VOID = 0;
/** \brief State value for parsing profile files
 */
const long PROCESS_CAM = 1;
/** \brief State value for parsing profile files
 */
const long PROCESS_LENS = 2;



/**
 * \brief Find the last occurance of str1 in str2
 * \return The position in str2 of the last occurance of str1.
 *         0 if not found
 * \param str2 The string to search
 * \param str1 The substring to look for in str1
 *
 *****************************************************/
long instrrev(char *str2, char *str1)
{
	int length = strlen(str2);
	if (length > 0)
	{
		char *start = str2;
		char *back = &str2[length];
		while((*(back-1) != *str1) && back != start)
		{
			--back;
			length--;
		}
	}
	return(length);
}




/**
 * \brief Trim white space from str
 * \return The input string (str) minus any white space
 *         from either end
 * \param str The input string
 *
 *****************************************************/
char *trim(char *str)
{
	int length = strlen(str);
	if (length > 0)
	{
		char *data = str;
     	char *front = str;
     	char *back = &str[length];
     	while(isspace(*front))
     		++front;
     	while(isspace(*(back-1)))
     		--back;
     	while(front < back)
     		*(data++) = *(front++);
     	*data = 0;
     }
     return(str);
}




/**
 * \brief Copy len characters from srcStr to destStr
 *        starting from offset.
 * \param destStr The destination string
 * \param srcStr The source string
 * \param offset The offset in srcStr from which to start copying
 * \param len The number of characters to copy to destStr
 *
 *****************************************************/
void substr(char *destStr, char *srcStr, int offset, int len)
{
	int i;
	for(i = 0; i < len && srcStr[offset + i] != '\0'; i++)
		destStr[i] = srcStr[i + offset];
	destStr[i] = '\0';
}





/**
 * \brief If key matches the beginning of str, copy the value
 *        (ie the rest of the string minus key) to the global.
 *        variable rhs.
 * \return 0 if key not found, 1 otherwise
 * \param str The string possibly containing key and value
 * \param key The key to look for
 *
 *****************************************************/
int lhs(char *str, char *key)
{
	int ret_value = 0;
	if (strncmp(str, key, strlen(key)) != 0)
    {
        ret_value = 0;
    } else {
	substr(rhs, str, strlen(key), (strlen(str)-strlen(key)));
    ret_value = 1;
    }
    return(ret_value);
}




/**
 * \brief Sort the global array of coefficients, coef[],
 *        within a region defined by lb and ub.
 * \param lb The lower bound (inclusive) for the sort region
 * \param ub The upper bound (inclusive) for the sort region
 *
 *****************************************************/
void sortCoefs(long lb, long ub)
{
    long i, j, t;
    CoefType c;
    
    // optimized bubble sort
    // t is an index to the last unsorted number
    t = ub;
    while (t > 0)
    {
        j = t;
        t = 0;
        for (i = lb; i < j; i++)
        {
            if (coef[i].f > coef[i + 1].f)
            {	// out of order, so bubble-down
                c = coef[i];
                coef[i] = coef[i + 1];
                coef[i + 1] = c;
                // remember point of last swap
                t = i;
            }
        }
    }
}




/**
 * \brief Insert a camera into the linked list of cameras
 *        such that cameras are alphabetized by make within
 *        models and models are in the same order as in 
 *        "profile.txt".
 * \param pCam The camera to insert
 *
 *****************************************************/
void insertCam(CamNode *pCam)
{
    CamNode *pCamTmp;

    if (pCamHdr == NULL)	//If first camera, set pCamHdr
    {
        pCamHdr = pCam;
    } else {				//Else find where to insert the new Camera Node
        pCamTmp = pCamHdr;
        while (1)
        {
            if (pCamTmp->nextCam == NULL)	//Only one node so far
            	break;
            if (pCamTmp->nextCam->menuMake > pCam->menuMake) 
            	break;
            pCamTmp = pCamTmp->nextCam;
        }
//Insert the new Camera Node
        if (pCamTmp->menuMake > pCam->menuMake)
        {
            // place before first make on list
            pCam->nextCam = pCamTmp;
            pCamHdr = pCam;
        } else {
            // insert after pCamTmp
            pCam->nextCam = pCamTmp->nextCam;
            pCamTmp->nextCam = pCam;
        }
    }
    //printf("Inserted camera: %-35s\n%", pCam->menuModel);
}




/**
 * \brief Insert a lens into the linked list of lenses
 *        such that lenses belonging to the same group
 *        are together.
 * \param pLns The lens to insert
 *
 *****************************************************/
void insertLns(LnsNode *pLns)
{
    LnsNode *pLnsTmp;
    long i;
    
    // adjust focal length so it looks like converter detected
    if (pLns->converterFactor != 1.0 && pLns->converterDetected == 0)
    {
        for (i = pLns->coefLB; i <= pLns->coefUB; i++)
        {
            coef[i].f = coef[i].f * pLns->converterFactor;
            pLns->converterDetected = 1;
        }
    }
    
    // find group/lens insertion point
    // find last group match and append
    if (pLnsHdr == NULL)
    {
        pLnsHdr = pLns;
    } else {
        // find first reference to group
        pLnsTmp = pLnsHdr;
        while (1)
        {
            if (pLnsTmp->nextLns == NULL)
            	break;
            if (pLnsTmp->group == pLns->group)
            	break;
            pLnsTmp = pLnsTmp->nextLns;
        }

        // find last reference to group
        if (pLnsTmp->group == pLns->group)
        {
            while (1)
            {
                if (pLnsTmp->nextLns == NULL)
                	break;
                if (pLnsTmp->nextLns->group != pLns->group)
                	break;
                pLnsTmp = pLnsTmp->nextLns;
            }
        }

        // insert after pLnsTmp
        pLns->nextLns = pLnsTmp->nextLns;
        pLnsTmp->nextLns = pLns;
    }
    //printf("Inserted Lens: %-40s %-10s\n", pLns->menuLens, pLns->group);
}




/**
 * \brief Utility function to print error messages from parsing
 *        the profile files.
 * \return The value of b passed in
 * \param b Indicates if an error message needs to be printed
 * \param s The string identifying a missing element in the profile file
 *
 *****************************************************/
int check(int b, char *s)
{
    if (b == 0)
        printf("File %s, line %d, missing: %s\n", fileName, line, s);

    return(b);
}




/**
 * \brief Function to parse profile files. The Lens, Camera, and
 *        group linked lists are created here, and the coef[] array
 *        is populated.
 * \return 0 if line parsed successfully, 1 otherwise
 * \param s A line from a profile file
 *
 *****************************************************/
long processDbFile(char *s)
{
	long ret_value=0;
	static long state;
    static CamNode *pCam;
    static LnsNode *pLns;  
    static int lnsMenuLens;
    static int lnsCalABC;
    static int lnsMultiplier;
    static int lnsGroup;
    static int lnsConverterFactor;
    static int lnsConverterDetected;
    
    static int camMenuMake;
    static int camMenuModel;
    static int camExifMake;
    static int camExifModel;
    static int camMultiplier;
    static int camGroup;

    if (strlen(s) == 0)
    	return(ret_value);
    if (strncmp(s, "#", 1) == 0)
    	return(0);

    if (lhs(s, "begin") == 1)
    {
        if (state != PROCESS_VOID)
        	goto lblInvalidStmt;
        if (strcmp(trim(rhs), "camera") == 0)
        {
            state = PROCESS_CAM;
            pCam = (CamNode *)malloc(sizeof(CamNode));
    		pCam->nextCam = NULL;
			pCam->group = NULL;
            camMenuMake = 0;
            camMenuModel = 0;
            camExifMake = 0;
            camExifModel = 0;
            camGroup = 0;
            camMultiplier = 0; 
        } else if (strcmp(trim(rhs), "lens") == 0)
        {
            state = PROCESS_LENS;
            pLns = (LnsNode *)malloc(sizeof(LnsNode));
            pLns->nextLns = NULL;
            pLns->coefLB = -1;
            pLns->coefUB = -1;
            pLns->converterFactor = 1.0;
            lnsMenuLens = 0;
            lnsGroup = 0;
            lnsMultiplier = 0;
            lnsCalABC = 0;
            lnsConverterFactor = 0;
            lnsConverterDetected = 0;
        }
   
    } else if (lhs(s, "end") == 1){
        if (state == PROCESS_CAM)
//Insert a camera node
        {
            if (check(camMenuMake, "menu_make") == 0) return(1);
            if (check(camMenuModel, "menu_model") == 0) return(1);
            if (check(camExifMake, "exif_make") == 0) return(1);
            if (check(camExifMake, "exif_model") == 0) return(1);
            if (check(camMultiplier, "multiplier") == 0) return(1);
            if (check(camGroup, "group") == 0) return(1);
            insertCam(pCam);
//Insert a lens node
        } else if (state == PROCESS_LENS){
            if (check(lnsMenuLens, "menu_lens") == 0) return(1);
            if (check(lnsCalABC, "cal_abc") == 0) return(1);
            if (check(lnsGroup, "group") == 0) return(1);
            if (check(lnsMultiplier, "multiplier") == 0) return(1);
            
            if (lnsConverterFactor == 1)
            {
                if (check(lnsConverterDetected, "converter_detected") == 0)
                    return(1);
            }
            sortCoefs(pLns->coefLB, pLns->coefUB);
            insertLns(pLns);
        } else {
            goto lblInvalidStmt;
        }
        
        state = PROCESS_VOID;

    } else if (lhs(s, "menu_make:") == 1) {
        if (state != PROCESS_CAM) goto lblInvalidStmt;
        strcpy(pCam->menuMake, trim(rhs));
        camMenuMake = 1;
            
    } else if (lhs(s, "menu_model:") == 1) {
        if (state != PROCESS_CAM) goto lblInvalidStmt;
        strcpy(pCam->menuModel, trim(rhs));
        camMenuModel = 1;

    } else if (lhs(s, "exif_make:") == 1) {
        if (state != PROCESS_CAM) goto lblInvalidStmt;
        strcpy(pCam->exifMake, trim(rhs));
        camExifMake = 1;
    
    } else if (lhs(s, "exif_model:") == 1) {
        if (state != PROCESS_CAM) goto lblInvalidStmt;
        strcpy(pCam->exifModel, trim(rhs));
        camExifModel = 1;
    
    } else if (lhs(s, "group:") == 1) {
        GrpNode *pGrp, *p;
        if (state == PROCESS_CAM)
        {
            pGrp = (GrpNode *)malloc(sizeof(GrpNode));
            strcpy(pGrp->name, trim(rhs));
            pGrp->nextGrp = NULL;

            // add new group to end of the groups list
            if (pCam->group == NULL)
            {
                pCam->group = pGrp;
            } else {
                p = pCam->group;
                while (p->nextGrp != NULL)
                    p = p->nextGrp;
                p->nextGrp = pGrp;
            }
            camGroup = 1;
        } else if (state == PROCESS_LENS) {
            strcpy(pLns->group, trim(rhs));
            lnsGroup = 1;
        } else {
            goto lblInvalidStmt;
        }

    } else if (lhs(s, "multiplier:") == 1) {
        if (state == PROCESS_CAM)
        {
            pCam->multiplier = (float)atof(rhs);
            camMultiplier = 1;
        } else if (state == PROCESS_LENS) {
            pLns->multiplier = (float)atof(rhs);
            lnsMultiplier = 1;
        } else {
            goto lblInvalidStmt;
        }
        
    } else if (lhs(s, "menu_lens:") == 1) {
        if (state != PROCESS_LENS) goto lblInvalidStmt;
        strcpy(pLns->menuLens, trim(rhs));
        lnsMenuLens = 1;
    
    } else if (lhs(s, "converter_factor:") == 1) {
        if (state != PROCESS_LENS) goto lblInvalidStmt;
        pLns->converterFactor = (float)atof(rhs);
        lnsConverterFactor = 1;
    
    } else if (lhs(s, "converter_detected:") == 1) {
        if (state != PROCESS_LENS) goto lblInvalidStmt;
        pLns->converterDetected = atol(rhs);
        lnsConverterDetected = 1;
    
    } else if (lhs(s, "cal_abc:") == 1) {
        if (state != PROCESS_LENS) goto lblInvalidStmt;
        if (pLns->coefLB == -1) pLns->coefLB = coefIndex;
        pLns->coefUB = coefIndex;
        sscanf(rhs, "%f %f %f %f", &(coef[coefIndex].f), &(coef[coefIndex].a),
        				&(coef[coefIndex].b), &(coef[coefIndex].c));
        coefIndex = coefIndex + 1;
        lnsCalABC = 1;
    } else {
        goto lblInvalidStmt;
    }

    return(ret_value);

lblInvalidStmt:
    printf( "File %s line %d invalid statement %s\n", fileName, line, s);
    return(1);
}




/**
 * \brief Parse the lines in the master profile file, "profile.txt".
 * \return 0 if line parsed successfully, 1 otherwise
 * \param s A line from "profile.txt"
 *
 *****************************************************/
long processDbIndex(char *s)
{
    if (strlen(s) == 0)
    	return(0);
    if (strncmp(s, "#", 1) == 0)
    	return(0);
    
    if (lhs(s, "version:") == 1)
    {
        // rhs contains version
        if (strcmp(trim(rhs), "4.2") != 0)
        {
            printf("Profile version 4.2 required, found version %s\n", rhs);
            return(1);
        }
    } else if (lhs(s, "file:") == 1) {
    	fileList[fileIndex] = (char *)malloc((strlen(rhs) + 1) * sizeof(char));
        strcpy(fileList[fileIndex], trim(rhs));
        fileIndex = fileIndex + 1;
    } else {
        printf("File %s, line %d, invalid statement: %s\n", fileName, line, s);
        return(1);
    }
    return(0);
}




/**
 * \brief Read lens calibration data from profile files.
 * \return 0 if successfull, 1 otherwise
 *
 *****************************************************/
long readDataBase()
{
    CamNode *pCam, *firstModel;
    LnsNode *pLns, *firstLns;
    char prevMake[100], prevModel[100], prevGroup[100];
	FILE *file;
	char dir[512];			//directory path to profile files
	char nextFile[512];
    char s[1000];			//holds input lines from profile files
    long i;
    long ret_value;
    int pathlen;
    struct stat file_info;

//Open profile.txt and read in the list of profile files
    fileIndex=0;
    // verify presence of profile.txt
    fileName = pref.profilePath;
    if (stat(fileName, &file_info) == 0)
    {
    	if (S_ISREG(file_info.st_mode) == 0)
    	{
    		printf("%s is not a profile file\n", fileName);
    		return(1);
    	}
    } else {
    	if (errno == ENOENT)
    	{
    		printf("%s does not exist\n", fileName);
    	}
    }
	file = fopen(fileName, "r");
    if (file == NULL)
    {
        printf("Unable to open profile file: %s\n", fileName);
        return(1);
    }
    // get profile files from profile.txt
    line = 0;
    while (fgets(s,1000,file)!=NULL)
    {
        line = line + 1;
        ret_value = processDbIndex(s);
        if (ret_value != 0)
        {
            fclose(file);
            return(ret_value);
        }
    }
    fclose(file);

//Read in each profile file
	/* Save the directory part of the profilePath in dir
	 * First look for the last '/'. If not found, look for
	 * the last '\'. If that's not found assume no path part
	 * in the profilePath specification (ie it is in the 
	 * current directory)
	*/
	pathlen = instrrev(pref.profilePath, "/");
	if (pathlen == 0)
		pathlen = instrrev(pref.profilePath, "\\");
	if (pathlen != 0)
	{
		strncpy(dir, pref.profilePath, pathlen);
		dir[pathlen] = 0;
	}

    for (i=0; i < fileIndex; i++)
    {
    	strcpy(nextFile, dir);
        strcat(nextFile, fileList[i]);
        fileName = nextFile;
        file = fopen(fileName, "r");
        if (file == NULL)
        {
            printf("Unable to open profile file: %s\n", fileName);
            return(1);
        }
        line = 0;
        while (fgets((s),1000,file)!=NULL)
        {
        	line = line + 1;
            ret_value = processDbFile(trim(s));
            if (ret_value != 0)
            {
                fclose(file);
                return(ret_value);
            }
        }
        fclose(file);
    }

//find first camera model
    prevMake[0] = 0;			//Set prevMake to NULL string
    pCam = pCamHdr;
    while (pCam != NULL)
    {
        if (pCam->menuMake != prevMake)
        {
            strcpy(prevMake, pCam->menuMake);
            prevModel[0] = 0;	//Set prevModel to NULL string
            firstModel = pCam;
        }
        if (strcmp(pCam->menuModel, prevModel) != 0)
            strcpy(prevModel, pCam->menuModel);

        pCam->firstModel = firstModel;
        pCam = pCam->nextCam;
    }
    
// find first lens
    pLns = pLnsHdr;
    prevGroup[0] = 0;			//Set prevGroup to NULL string
    while (pLns != NULL)
    {
        if (strcmp(pLns->group, prevGroup) != 0)
        {
            strcpy(prevGroup, pLns->group);
            firstLns = pLns;
        }
        pLns->firstLns = firstLns;
        pLns = pLns->nextLns;
        }
    return(0);
}




/**
 * \brief Look for a camera in the linked list of supported cameras.
 * \return pointer to camera node if successfull, NULL otherwise
 * \param exifMake The camera make as found in the jpeg EXIF data
 * \param exifModel The camera model as found in the jpeg EXIF data
 *
 *****************************************************/
 CamNode *findCamera(char *exifMake, char *exifModel)
 {
	 CamNode *pCurCam;

	 pCurCam = pCamHdr;
	 while((pCurCam != NULL) && ((strcmp(pCurCam->exifModel, exifModel) != 0) ||
						(strcmp(pCurCam->exifMake, exifMake) != 0)))
		 pCurCam = pCurCam->nextCam;
	 return(pCurCam);
 }
 



/**
 * \brief Look for a lens in the list of lens groups compatable with
 *        the camera.
 * \return 0 if unseccessful, 1 otherwise
 * \param thisLens The lens to look for
 * \param thisCamera The camera model as found in the jpeg EXIF data
 *
 *****************************************************/
 int inGroups(LnsNode *thisLens, CamNode *thisCamera)
 {
	 GrpNode *pCurGroup;

	 pCurGroup = thisCamera->group;
	 while(pCurGroup != NULL && (strcmp(pCurGroup->name, thisLens->group) != 0))
		 pCurGroup = pCurGroup->nextGrp;
	 if(pCurGroup == NULL)
		 return(0);
	 else
		 return(1);
 }
 


/**
 * \brief Look for a lens in the linked list of supported lenses.  
 *        The lens is checked to make sure it has been calibrated
 *        for the camera passed in.
 * \return pointer to the lens node if successful, NULL otherwise
 * \param lens The lens to look for
 * \param camera The camera model as found in the jpeg EXIF data
 *
 *****************************************************/
 LnsNode *findLens(char *lens, CamNode *camera)
 {
	 LnsNode *pCurLens;

	 pCurLens = pLnsHdr;
	 while(pCurLens != NULL)
	 {
		 if(inGroups(pCurLens, camera) == 1)			//If this lens supports the current camera
			 if(strcmp(pCurLens->menuLens, lens) == 0)	//See if this lens is the one the user used
				 return(pCurLens);
		 pCurLens = pCurLens->nextLns;
	 }
	 return(NULL);										//If here, lens not found (pCurLens is NULL)
 }

 
 

/**
 * \brief Create a linked list of all calibrated lenses for
 *        the input camera.
 * \return pointer to the first lens node if lenses found,
 *         NULL otherwise
 * \param camera A pointer to the camera node in the linked list
 *
 *****************************************************/
 LnsNode *findLenses(CamNode *camera)
 {
	 LnsNode *pDBLens;
	 LnsNode *pCamLens;
	 LnsNode *pFirstLens = NULL;

	 pDBLens = pLnsHdr;
	 while(pDBLens != NULL)
	 {
		 if(inGroups(pDBLens, camera) == 1)			//If this lens supports the current camera
		 {
			 if(pFirstLens == NULL)
			 {
				 pCamLens = (LnsNode *)malloc(sizeof(LnsNode));
				 pCamLens->nextLns = NULL;
				 pCamLens->firstLns = pDBLens->firstLns;
				 pCamLens->numLens = pDBLens->numLens;
				 strcpy(pCamLens->group, pDBLens->group);
				 strcpy(pCamLens->menuLens, pDBLens->menuLens);
				 pCamLens->converterFactor = pDBLens->converterFactor;
				 pCamLens->converterDetected = pDBLens->converterDetected;
				 pCamLens->coefLB = pDBLens->coefLB;
				 pCamLens->coefUB = pDBLens->coefUB;
				 pCamLens->multiplier = pDBLens->multiplier;
				 pFirstLens = pCamLens;
			 } else {
				 pCamLens->nextLns = (LnsNode *)malloc(sizeof(LnsNode));
				 pCamLens = pCamLens->nextLns;
				 pCamLens->nextLns = NULL;
				 pCamLens->firstLns = pDBLens->firstLns;
				 pCamLens->numLens = pDBLens->numLens;
				 strcpy(pCamLens->group, pDBLens->group);
				 strcpy(pCamLens->menuLens, pDBLens->menuLens);
				 pCamLens->converterFactor = pDBLens->converterFactor;
				 pCamLens->converterDetected = pDBLens->converterDetected;
				 pCamLens->coefLB = pDBLens->coefLB;
				 pCamLens->coefUB = pDBLens->coefUB;
				 pCamLens->multiplier = pDBLens->multiplier;
			 }
		 }
		 pDBLens = pDBLens->nextLns;
	 }
	 return(pFirstLens);
 }
